import aio_pika
import aio_pika.connection
import asyncio
import contextvars
import inspect
import logging
import pydantic
import socket
import time
import uuid

from aio_pika import IncomingMessage, Message
from aio_pika.abc import (
    AbstractChannel,
    AbstractExchange,
    AbstractRobustConnection,
    AbstractQueue,
)

from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast

from . import errors
from . import metrics
from . import utils
from .connection_manager import get_connection
from .models import Request, Response
from .types import Priority, RequestContext


logger = logging.getLogger(__name__)

request_context: contextvars.ContextVar[Optional[RequestContext]] = (
    contextvars.ContextVar("request_context", default=None)
)


ExceptionPredicate = Callable[[Exception], bool]

IsinstanceType = Union[Type[Exception], Tuple[Type[Exception], ...]]


class RpcServer:
    def __init__(self, service_name: str) -> None:
        self._service_name = utils.clean_service_name(service_name)
        self._app_id = f"{self._service_name}@{socket.gethostname()}"

        self._handlers: Dict[str, Callable] = {}
        self._retryable_errors_predicates: List[ExceptionPredicate] = []
        self._retryable_errors_configured = False

        self._reset()

        self.bind(self._ping)
        self.bind(self._help)

    def _reset(self) -> None:
        self._connection: Optional[AbstractRobustConnection] = None
        self._channel: Optional[AbstractChannel] = None
        self._exchange: Optional[AbstractExchange] = None
        self._queue: Optional[AbstractQueue] = None

    def bind(
        self, handler: Optional[Callable] = None, *, name: Optional[str] = None
    ) -> Callable:
        """
        Binds to be exposed handlers (functions) to RPC server instance:

        @server.bind
        def add(*, a: float, b: float) -> float:
            ...

        # or

        def sub(*, a: float, b: float) -> float:
            ...

        server.bind(sub)

        # or

        @server.bind(name="mul")
        def multiply(*, a: float, b: float) -> float:
            ...

        """

        def decorator(handler: Callable) -> Callable:
            self._bind_handler(handler=handler, name=name)
            return handler

        if handler:
            return decorator(handler)
        else:
            return decorator

    def _bind_handler(self, *, handler: Callable, name: Optional[str] = None) -> None:
        handler_name = name or handler.__name__
        if handler_name in self._handlers:
            raise KeyError(f"Handler {handler_name} already defined")

        if hasattr(handler, "raw_function"):
            # Unwrap `pydantic.validate_call` decorator
            handler = handler.raw_function

        utils.assert_handler_is_coroutine(handler)
        utils.assert_handler_kwargs_only(handler)

        self._handlers[handler_name] = pydantic.validate_call(validate_return=True)(
            handler
        )

    @pydantic.validate_call
    def retryable_if(
        self,
        error_types: Optional[Union[IsinstanceType, ExceptionPredicate]] = None,
        predicate: Optional[ExceptionPredicate] = None,
    ) -> None:
        """
        Configure which exceptions to mark as retryable.
        `error_types` could be an Exception type, a tuple of of Exception types or a Predicate.
        Predicate could be a callable callable that returns boolean

        The error must match *any* `error_types` *and* predicate same time.

        Examples:

        rpc_server.retryable_if(ConnectionError)

        rpc_server.retryable_if( (ConnectionRefusedError, ConnectionResetError) )

        rpc_server.retryable_if(
            httpx.HTTPStatusError,
            lambda error: 500 <= error.response.status_code < 600
        )

        rpc_server.retryable_if(lambda error: "retry" in str(error))
        """
        if not (error_types or predicate):
            return

        if (
            predicate
            and isinstance(predicate, type)
            and issubclass(predicate, Exception)
        ):
            raise TypeError("`predicate` must be function")

        if not (isinstance(error_types, type) or isinstance(error_types, tuple)):
            if predicate:
                raise TypeError("Only one predicate is allowed")
            else:
                predicate = error_types
                error_types = None

        is_retryable: ExceptionPredicate

        if error_types and predicate:

            def is_retryable(error: Exception) -> bool:
                return isinstance(error, error_types) and predicate(error)

        elif error_types:

            def is_retryable(error: Exception) -> bool:
                return isinstance(error, error_types)

        else:
            assert predicate
            is_retryable = predicate

        self._retryable_errors_predicates.append(is_retryable)

    def configure_default_retryable_errors(self) -> None:
        if self._retryable_errors_configured:
            return
        self._retryable_errors_configured = True

        try:
            import httpx
        except ImportError:
            pass
        else:
            self.retryable_if(
                httpx.HTTPStatusError,
                lambda error: 500 <= error.response.status_code < 600,  # type: ignore[attr-defined]
            )

        try:
            import requests
        except ImportError:
            pass
        else:
            self.retryable_if(
                requests.HTTPError,
                lambda error: 500 <= error.response.status_code < 600,  # type: ignore[attr-defined]
            )

        try:
            from torch.cuda import OutOfMemoryError  # type: ignore[import-not-found]
        except ImportError:
            pass
        else:
            self.retryable_if(OutOfMemoryError)

        # https://grpc.github.io/grpc/core/md_doc_statuscodes.html
        GRPC_RETRYABLE_CODES = {
            "UNKNOWN",
            "DEADLINE_EXCEEDED",
            "RESOURCE_EXHAUSTED",
            "INTERNAL",
            "UNAVAILABLE",
        }

        try:
            import grpc
        except ImportError:
            pass
        else:

            def if_retryable_grpc(error: Exception) -> bool:
                if not hasattr(error, "code"):
                    return False

                status_code = error.code()

                return status_code.name in GRPC_RETRYABLE_CODES

            self.retryable_if(grpc.RpcError, if_retryable_grpc)

        try:
            import grpclib
        except ImportError:
            pass
        else:
            self.retryable_if(
                grpclib.GRPCError,
                lambda error: error.status.name in GRPC_RETRYABLE_CODES,  # type: ignore[attr-defined]
            )

    async def start(self, connection_string: str, prefetch_count: int = 1) -> None:
        self.configure_default_retryable_errors()

        if self._connection:
            raise RuntimeError("Already started")

        connection = self._connection = await get_connection(
            connection_string, self._service_name
        )
        channel = self._channel = await connection.channel()
        await channel.set_qos(prefetch_count=prefetch_count)

        exchange = self._exchange = await channel.declare_exchange(
            name=f"service.{self._service_name}",
            type=aio_pika.ExchangeType.TOPIC,
            durable=True,
        )

        queue = self._queue = await channel.declare_queue(
            name=f"service.{self._service_name}",
            durable=True,
            arguments={
                "x-max-priority": max(Priority),
                "x-dead-letter-exchange": f"service.{self._service_name}.dlx",
            },
        )

        await queue.bind(exchange, "#")
        await queue.consume(utils.fix_message_timestamp(self._on_message))

        logger.info("RPC server is consuming messages from `%s`", queue)

    async def stop(self) -> None:
        channel = self._channel
        if channel:
            self._reset()
            await channel.close()

    async def serve_forever(
        self, connection_string: str, prefetch_count: int = 1
    ) -> None:
        await self.start(connection_string, prefetch_count=prefetch_count)
        try:
            await asyncio.Future()
        finally:
            await self.stop()

    async def _on_message(self, message: IncomingMessage) -> None:
        consumed_timestamp = time.time()

        if not message.message_id:
            message.message_id = str(uuid.uuid1(node=0))

        logger.debug(
            "[MID:%s] Got `%s` message from `%s` service",
            message.message_id,
            message.type or "[untyped]",
            message.app_id or "unknown",
        )

        if message.type != "rpc.request":
            logger.warning(
                "[MID:%s] Unexpected message type: `%s`",
                message.message_id,
                message.type,
            )
            await message.reject()
            metrics.message_consumed(message, error="WRONG_MESSAGE_TYPE")
            return

        published_timestamp = None
        if message.timestamp:
            published_timestamp = message.timestamp.timestamp()

        deadline = None
        if published_timestamp and message.expiration:
            deadline = published_timestamp + cast(float, message.expiration)

        request_context.set(
            RequestContext(
                deadline=deadline,
                priority=Priority(message.priority or 0),
            )
        )

        try:
            response = await asyncio.wait_for(
                self._process_message(message, consumed_timestamp, published_timestamp),
                timeout=(deadline - time.time()) if deadline else None,
            )
        except asyncio.TimeoutError:
            logger.info(
                "[MID:%s] [CID:%s] Execution of `%s` from `%s` service timed out",  # noqa: E501
                message.message_id,
                message.correlation_id,
                message.type,
                message.app_id or "unknown",
            )

            await message.ack()
            metrics.message_consumed(
                message, routing_key=message.routing_key, error="TIMEOUT"
            )
            return

        if message.reply_to:
            expiration = None
            if deadline:
                expiration = deadline - time.time()

            if expiration is None or expiration > 0:
                response_message = Message(
                    body=response.model_dump_msgpack(),
                    content_type="application/msgpack",
                    delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
                    priority=message.priority,
                    correlation_id=message.correlation_id,
                    expiration=expiration,
                    message_id=str(uuid.uuid1()),
                    timestamp=time.time(),
                    type="rpc.response",
                    app_id=self._app_id,
                )

                await message.channel.basic_publish(
                    body=response_message.body,
                    routing_key=message.reply_to,
                    properties=response_message.properties,
                )
                metrics.message_published(
                    exchange="",
                    routing_key="",  # random, causes high cardinality metric
                    type=response_message.type or "",
                    priority=response_message.priority or 0,
                    payload_size=response_message.body_size,
                )

        await message.ack()
        metrics.message_consumed(message, routing_key=message.routing_key)

        logger.debug(
            "[MID:%s] [CID:%s] Handled `%s` from `%s` service, success: %s",  # noqa: E501
            message.message_id,
            message.correlation_id,
            message.type,
            message.app_id or "unknown",
            response.error is None,
        )

    @Response.wrap_errors
    async def _process_message(
        self,
        message: IncomingMessage,
        consumed_timestamp: float,
        published_timestamp: Optional[float],
    ) -> Response:
        request = Request.model_validate_message(message)

        timings = request.timings
        queued_for = None
        if published_timestamp is not None:
            timings.append(("queue.published", published_timestamp))
            queued_for = consumed_timestamp - published_timestamp

        timings.append((f"queue.consumed by {self._app_id}", consumed_timestamp))

        with metrics.server_requests_in_progress.labels(
            method=request.method_name
        ).track_inprogress():
            elapsed = -time.perf_counter()
            response = await self._process_request(request)
            elapsed += time.perf_counter()
        timings.append(("execution.completed", consumed_timestamp + elapsed))

        response.context = request.context
        response.timings = timings

        metrics.server_request_processed(
            priority=message.priority or 0,
            method=request.method_name,
            error=response.error,
            queued_for=queued_for or 0,
            elapsed=elapsed,
        )

        logger.info(
            "[MID:%s] Processed `%s` from `%s` service to method `%s`. In queue: %ims, execution: %ims, success: %s%s",
            message.message_id,
            message.type,
            message.app_id or "unknown",
            request.method_name,
            (None if queued_for is None else max(queued_for * 1000, 0)),
            elapsed * 1000,
            response.error is None,
            (f", error: {response.error.message}" if response.error else ""),
        )

        return response

    @Response.wrap_errors
    async def _process_request(self, request: Request) -> Response:
        handler = self._handlers.get(request.method_name)
        if handler is None:
            raise errors.UnknownMethodError(request.method_name)

        try:
            result = await handler(**request.arguments)
        except Exception as error:
            logger.debug(
                f"Exception while processing '{request.method_name}'", exc_info=True
            )
            error_type: Type[errors.RpcError] = errors.ExecutionError
            if any(predicate(error) for predicate in self._retryable_errors_predicates):
                error_type = errors.RetryableExecutionError

            raise error_type(repr(error))

        if isinstance(result, pydantic.BaseModel):
            result = result.model_dump()

        return Response(result=result)

    async def _ping(self) -> None:
        return None

    async def _help(self) -> Dict[str, Any]:
        """
        Returns methods list
        """
        return {
            "methods": [
                {
                    "method_name": method_name,
                    "signature": f"{method_name}{inspect.signature(handler)}",
                    "description": handler.__doc__,
                }
                for method_name, handler in self._handlers.items()
            ]
        }



import pydantic

from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, Any, Optional, TypeVar


__all__ = [
    "AbstractOptions",
    "Priority",
    "RequestContext",
]


TAbstractOptions = TypeVar("TAbstractOptions", bound="AbstractOptions")


class AbstractOptions(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid", frozen=True)

    def __init__(self, *args: "AbstractOptions", **kwargs: Dict[str, Any]) -> None:
        if args:
            merged = {}
            self_class = type(self)
            for options in args:
                if not isinstance(options, self_class):
                    raise TypeError(
                        f"Positional arguments must be {self_class} instances"
                    )

                merged.update(options.model_dump(exclude_unset=True))
            merged.update(**kwargs)

            kwargs = merged

        return super().__init__(**kwargs)

    def update(
        self: TAbstractOptions, *args: "TAbstractOptions", **kwargs: Any
    ) -> "TAbstractOptions":
        if not args and not kwargs:
            return self

        return self.__class__(self, *args, **kwargs)


class Priority(IntEnum):
    LOW = 0
    NORMAL = 1
    DEFAULT = 1
    HIGH = 2
    INTERACTIVE = 3
    SYSTEM = 4


@dataclass(frozen=True)
class RequestContext:
    deadline: Optional[float]
    priority: Priority



import importlib.metadata

from .models import ResponseError

from aio_pika import IncomingMessage
from prometheus_client import Counter, Gauge, Histogram, Info, Summary
from typing import Optional


application_info = Info("application", "Application info")


def set_application_info(**kwargs: str) -> None:
    mertics_package = __package__.split(".")[0]
    application_info.info(
        {
            "mertics_package": mertics_package,
            "mertics_package_version": importlib.metadata.version(mertics_package),
            **kwargs,
        }
    )


set_application_info()


messages_published_count = Counter(
    name="queue_rpc_messages_published_total",
    documentation="Total number of published messages",
    labelnames=["exchange", "routing_key", "type", "priority"],
)

messages_published_payload_size = Summary(
    name="queue_rpc_messages_published_payload_size_bytes",
    documentation="Payload size of published messages",
    labelnames=["exchange", "routing_key", "type"],
)


def message_published(
    *, exchange: str, routing_key: str, type: str, priority: int, payload_size: int
) -> None:
    labels = dict(
        exchange=exchange,
        routing_key=routing_key,
        type=type,
    )
    messages_published_count.labels(priority=priority, **labels).inc()
    messages_published_payload_size.labels(**labels).observe(payload_size)


messages_consumed_count = Counter(
    name="queue_rpc_messages_consumed_total",
    documentation="Total number of consumed messages",
    labelnames=["exchange", "routing_key", "redelivered", "type", "priority", "error"],
)

messages_consumed_payload_size = Summary(
    name="queue_rpc_messages_consumed_payload_size_bytes",
    documentation="Payload size of consumed messages",
    labelnames=["exchange", "routing_key", "redelivered", "type"],
)


def message_consumed(
    message: IncomingMessage,
    *,
    routing_key: Optional[str] = None,
    error: Optional[str] = None,
) -> None:
    labels = dict(
        exchange=message.exchange,
        routing_key=routing_key or "",
        redelivered=message.redelivered,
        type=message.type,
    )
    messages_consumed_count.labels(
        priority=message.priority, error=error or "", **labels
    ).inc()
    messages_consumed_payload_size.labels(**labels).observe(message.body_size)


server_requests_count = Counter(
    name="queue_rpc_server_requests_total",
    documentation="Total number of requests",
    labelnames=["priority", "method", "error"],
)

server_requests_in_progress = Gauge(
    name="queue_rpc_server_requests_in_progress",
    documentation="Number of requests in progress",
    labelnames=["method"],
)

server_request_waiting_time = Histogram(
    name="queue_rpc_server_request_waiting_seconds",
    documentation="Time request spent in queue",
    labelnames=["priority", "method"],
    buckets=(
        0.1,
        0.3,
        0.5,
        1,
        3,
        5,
        10,
        30,
        1 * 60,
        3 * 60,
        5 * 60,
        10 * 60,
        30 * 60,
        1 * 3600,
        3 * 3600,
        5 * 3600,
        10 * 3600,
        30 * 3600,
    ),
)

server_request_processing_time = Histogram(
    name="queue_rpc_server_request_processing_seconds",
    documentation="Time spent processing request",
    labelnames=["method"],
    buckets=(
        0.1,
        0.25,
        0.5,
        0.75,
        1.0,
        2.5,
        5.0,
        7.5,
        10,
        25,
        50,
        75,
        100,
        250,
        500,
        750,
        1000,
        2500,
    ),
)


def server_request_processed(
    priority: int,
    method: str,
    error: Optional[ResponseError],
    queued_for: float,
    elapsed: float,
) -> None:
    server_requests_count.labels(
        priority=priority,
        method=method,
        error=error and error.type or "",
    ).inc()

    if queued_for is not None:
        server_request_waiting_time.labels(
            priority=priority,
            method=method,
        ).observe(queued_for)

    server_request_processing_time.labels(
        method=method,
    ).observe(elapsed)


client_requests_count = Counter(
    name="queue_rpc_client_requests_total",
    documentation="Total number of requests",
    labelnames=["priority", "service", "method"],
)

client_requests_in_progress = Gauge(
    name="queue_rpc_client_requests_in_progress",
    documentation="Number of requests in progress",
    labelnames=["service", "method"],
)

client_request_time = Histogram(
    name="queue_rpc_client_request_seconds",
    documentation="Time request took",
    labelnames=["priority", "service", "method"],
    buckets=(
        0.1,
        0.3,
        0.5,
        1,
        3,
        5,
        10,
        30,
        1 * 60,
        3 * 60,
        5 * 60,
        10 * 60,
        30 * 60,
        1 * 3600,
        3 * 3600,
        5 * 3600,
        10 * 3600,
        30 * 3600,
    ),
)


def client_request_processed(
    priority: int, service: str, method: str, elapsed: float
) -> None:
    labels = dict(
        priority=int(priority or 0),
        service=service,
        method=method,
    )

    client_requests_count.labels(**labels).inc()

    client_request_time.labels(**labels).observe(elapsed)



import aio_pika
import aio_pika.exceptions
import asyncio
import logging
import socket
import tenacity
import time
import uuid
import weakref

from aio_pika import IncomingMessage, Message
from aio_pika.abc import (
    AbstractChannel,
    AbstractExchange,
    AbstractRobustConnection,
    AbstractQueue,
)
from typing import Any, Awaitable, Dict, Optional, Union

from . import errors
from . import metrics
from .connection_manager import get_connection
from .models import Request, Response
from .server import request_context
from .types import AbstractOptions, Priority
from .utils import clean_service_name, fix_message_timestamp


logger = logging.getLogger(__name__)


class Options(AbstractOptions):
    timeout: Optional[float] = 60
    priority: Optional[Priority] = None
    no_wait: bool = False
    retry: Union[bool, Dict] = False


DEFAULT_RETRY_OPTIONS: Dict[str, Any] = {
    "wait": tenacity.wait_fixed(1) + tenacity.wait_random_exponential(max=60),
    "retry": tenacity.retry_if_exception_type(errors.RetryableError),
    "reraise": True,
}


class RpcClient:
    def __init__(
        self, client_name: str, connection_string: str, *args: Options, **kwargs: Any
    ) -> None:
        self._client_name = clean_service_name(client_name)
        self._connection_string = connection_string
        self._app_id = f"{self._client_name}@{socket.gethostname()}"

        self._options = Options(*args, **kwargs)
        self._client_lock = asyncio.Lock()
        self._declare_exchange_lock = asyncio.Lock()

        self._reset()

    def _reset(self) -> None:
        self._connection: Optional[AbstractRobustConnection] = None
        self._channel: Optional[AbstractChannel] = None
        self._reply_queue: Optional[AbstractQueue] = None

        self._exchanges: Dict[str, AbstractExchange] = {}
        self._futures: weakref.WeakValueDictionary[str, asyncio.Future] = (
            weakref.WeakValueDictionary()
        )

    def get_service(
        self, service_name: str, *args: Options, **kwargs: Dict[str, Any]
    ) -> "RpcService":
        options = self._options.update(*args, **kwargs)
        return RpcService(self, clean_service_name(service_name), options)

    __getitem__ = __getattr__ = get_service

    async def _ensure_connection(self) -> None:
        async with self._client_lock:
            if self._connection and self._channel:
                return

            self._connection = await get_connection(
                self._connection_string, self._client_name
            )
            self._channel = await self._connection.channel(on_return_raises=True)

    async def _ensure_reply_queue(self) -> None:
        await self._ensure_connection()

        assert self._channel

        async with self._client_lock:
            if self._reply_queue:
                return

            queue_name = f"client.{self._client_name}.{uuid.uuid4().hex}"
            queue = await self._channel.declare_queue(
                name=queue_name,
                exclusive=True,
                auto_delete=True,
            )
            await queue.consume(fix_message_timestamp(self._on_message))

            self._reply_queue = queue

    async def stop(self) -> None:
        channel = self._channel
        if channel:
            self._reset()
            await channel.close()

    def call(
        self,
        *,
        service_name: str,
        method_name: str,
        arguments: Optional[Dict[str, Any]] = None,
        options: Union[None, Options, dict] = None,
    ) -> Awaitable[Any]:
        if not options:
            options = self._options

        elif isinstance(options, Options):
            options = self._options.update(options)

        else:
            options = self._options.update(**options)

        if options.retry or isinstance(options.retry, dict):
            retry_options = DEFAULT_RETRY_OPTIONS
            if isinstance(options.retry, dict):
                retry_options = dict(retry_options, **options.retry)

            retrying = tenacity.AsyncRetrying(**retry_options)
            do_call = retrying.wraps(self._call)
        else:
            do_call = self._call

        return do_call(
            service_name=service_name,
            method_name=method_name,
            arguments=arguments or {},
            options=options,
        )

    async def _call(
        self,
        *,
        service_name: str,
        method_name: str,
        arguments: Dict[str, Any],
        options: Options,
    ) -> Any:
        current_request_context = request_context.get()
        if current_request_context:
            if options.priority is None:
                options = options.update(priority=current_request_context.priority)

            if current_request_context.deadline is not None:
                timeout = current_request_context.deadline - time.time()
                if options.timeout is not None:
                    timeout = min(options.timeout, timeout)

                options = options.update(timeout=timeout)

        if options.priority is None:
            options = options.update(priority=Priority.DEFAULT)

        if options.timeout:
            deadline = time.monotonic() + options.timeout
            waiter = lambda coro: asyncio.wait_for(  # noqa: E731
                coro, timeout=(deadline - time.monotonic())
            )
        else:
            waiter = lambda coro: coro  # noqa: E731

        request = Request(method_name=method_name, arguments=arguments)

        if options.no_wait:
            correlation_id = None
            future = None
        else:
            correlation_id = str(uuid.uuid1())
            future = asyncio.get_running_loop().create_future()
            self._futures[correlation_id] = future

        with metrics.client_requests_in_progress.labels(
            service=service_name,
            method=method_name,
        ).track_inprogress():
            elapsed = -time.perf_counter()
            await waiter(
                self._send_request(service_name, request, options, correlation_id)
            )

            try:
                if future:
                    return await waiter(future)
            except asyncio.CancelledError:
                logger.warning(
                    "Wait was cancelled but not the request itself. "
                    "Pass `timeout` option instead of using `asyncio.wait_for` or similar"
                )
                raise
            finally:
                elapsed += time.perf_counter()

                metrics.client_request_processed(
                    priority=options.priority or 0,
                    service=service_name,
                    method=method_name,
                    elapsed=elapsed,
                )

    async def _send_request(
        self,
        service_name: str,
        request: Request,
        options: Options,
        correlation_id: Optional[str] = None,
    ) -> None:
        await self._ensure_connection()

        assert self._channel

        request_message = Message(
            body=request.model_dump_msgpack(),
            content_type="application/msgpack",
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
            priority=options.priority,
            expiration=options.timeout or None,
            message_id=str(uuid.uuid1()),
            timestamp=time.time(),
            type="rpc.request",
            app_id=self._app_id,
        )

        if correlation_id:
            await self._ensure_reply_queue()
            assert self._reply_queue
            request_message.correlation_id = correlation_id
            request_message.reply_to = self._reply_queue.name

        async with self._declare_exchange_lock:
            if service_name not in self._exchanges:
                try:
                    self._exchanges[service_name] = await self._channel.get_exchange(
                        f"service.{service_name}"
                    )
                except aio_pika.exceptions.ChannelNotFoundEntity:
                    raise errors.UnknownServiceError(
                        f"Exchange `{service_name}` was not found"
                    )

        try:
            routing_key = f"method.{request.method_name}"
            await self._exchanges[service_name].publish(
                message=request_message,
                routing_key=routing_key,
            )
            metrics.message_published(
                exchange=self._exchanges[service_name].name,
                routing_key=routing_key,
                type=request_message.type or "",
                priority=request_message.priority or 0,
                payload_size=request_message.body_size,
            )

        except aio_pika.exceptions.PublishError:
            raise errors.UnknownServiceError("Request was not delivered to a queue")

    async def _on_message(self, message: IncomingMessage) -> None:
        if not message.message_id:
            message.message_id = str(uuid.uuid1(node=0))

        if message.type != "rpc.response":
            logger.warning(
                "[MID:%s] Unexpected message type: `%s`",
                message.message_id,
                message.type,
            )
            await message.reject()
            metrics.message_consumed(message, error="WRONG_MESSAGE_TYPE")
            return

        if not message.correlation_id:
            logger.warning("[MID:%s] `correlation_id` is not set", message.message_id)
            await message.reject()
            metrics.message_consumed(message, error="NO_CORRELATION_ID")
            return

        future = self._futures.pop(message.correlation_id, None)
        if not future or future.done():
            logger.warning(
                "[MID:%s] [CID:%s] Response is not awaited (too late or duplicate)",
                message.message_id,
                message.correlation_id,
            )
            await message.reject()
            metrics.message_consumed(message, error="UNKNOWN_CORRELATION_ID")
            return

        try:
            future.set_result(await self._process_message(message))
        except Exception as error:
            future.set_exception(error)

        await message.ack()
        metrics.message_consumed(message)

        logger.debug(
            "[MID:%s] [CID:%s] Got `%s` from `%s` service",
            message.message_id,
            message.correlation_id,
            message.type,
            message.app_id or "unknown",
        )

    async def _process_message(self, message: IncomingMessage) -> Any:
        response = Response.model_validate_message(message)

        if response.error:
            error_class = getattr(errors, response.error.type, errors.UnknownError)
            raise error_class(response.error.message)

        return response.result


class RpcService:
    def __init__(self, client: RpcClient, service_name: str, options: Options):
        self._client = client
        self._service_name = service_name
        self._options = options

    def get_method(
        self, method_name: str, *args: Options, **kwargs: Dict[str, Any]
    ) -> "RpcMethod":
        options = self._options.update(*args, **kwargs)
        return RpcMethod(self._client, self._service_name, method_name, options)

    __getitem__ = __getattr__ = get_method

    def __repr__(self) -> str:
        class_ = self.__class__
        return f"<{class_.__qualname__} `{self._service_name}`>"


class RpcMethod:
    def __init__(
        self, client: RpcClient, service_name: str, method_name: str, options: Options
    ) -> None:
        self._client = client
        self._service_name = service_name
        self._method_name = method_name
        self._options = options

    def __repr__(self) -> str:
        class_ = self.__class__
        return (
            f"<{class_.__qualname__} `{self._method_name}` "
            "of service `{self._service_name}`>"
        )

    def __call__(self, *args: Options, **kwargs: Dict[str, Any]) -> Any:
        options = self._options.update(*args)
        return self._client.call(
            service_name=self._service_name,
            method_name=self._method_name,
            arguments=kwargs,
            options=options,
        )
